import string
from typing import Any, AsyncIterable

from v6d2ctx.context import *
from v6d3music.utils.assert_admin import *
from v6d3music.utils.effects_for_preset import *
from v6d3music.utils.entries_for_url import *
from v6d3music.utils.options_for_effects import *
from v6d3music.utils.presets import *
from v6d3music.utils.sparq import *

__all__ = ('InfoCtx', 'BoundCtx', 'UrlCtx', 'ArgCtx',)


class InfoCtx:
    def __init__(
        self, info: dict[str, Any], effects: str | None, already_read: int, tor: bool, ignore: bool
    ) -> None:
        self.info = info
        self.effects = effects
        self.already_read = already_read
        self.tor = tor
        self.ignore = ignore

    def bind(self, ctx: Context) -> 'BoundCtx':
        return BoundCtx(self, ctx)


class BoundCtx:
    def __init__(self, it: InfoCtx, ctx: Context, /) -> None:
        if ctx.member is None:
            raise Explicit('not in a guild')
        self.member = ctx.member
        self.ctx = ctx
        self.url = it.info['url']
        self.description = f'{escape(it.info.get("title", "unknown"))} `Rby` {ctx.member}'
        self.tor = it.tor
        self.effects = it.effects
        self.already_read = it.already_read
        self.options = self._options()

    def _options(self) -> str | None:
        assert self.ctx.member is not None
        if self.effects:
            if self.effects not in allowed_effects:
                assert_admin(self.ctx.member)
            if not set(self.effects) <= set(string.ascii_letters + string.digits + '*,=+-/()|.^:_'):
                raise Explicit('malformed effects')
            return options_for_effects(self.effects)
        else:
            return None


class UrlCtx:
    def __init__(self, url: str, effects: str | None) -> None:
        self.url = url
        self.effects = effects
        self.already_read = 0
        self.tor = False
        self.ignore = False

    async def entries(self) -> AsyncIterable[InfoCtx]:
        try:
            async for info in entries_for_url(self.url, self.tor):
                yield InfoCtx(info, self.effects, self.already_read, self.tor, self.ignore)
        except Exception:
            if not self.ignore:
                raise


class ArgCtx:
    def __init__(self, default_effects: str | None, args: list[str]) -> None:
        self.sources: list[UrlCtx] = []
        while args:
            match args:
                case [url, '-', effects, *args]:
                    pass
                case [url, '+', preset, *args]:
                    effects = effects_for_preset(preset)
                case [url, *args]:
                    effects = default_effects
                case _:
                    raise RuntimeError
            if url in presets:
                raise Explicit('expected url, got preset. maybe you are missing `+`?')
            if url in {'+', '-'}:
                raise Explicit('expected url, got `+` or `-`. maybe you tried to use multiple effects?')
            if url.startswith('+') or url.startswith('-"') or url.startswith('-\''):
                raise Explicit('expected url, got `+` or `-"` or `-\'`. maybe you forgot to separate control symbol from the effects?')
            ctx = UrlCtx(url, effects)
            seconds = 0
            match args:
                case [h, m, s, *args] if h.isdecimal() and m.isdecimal() and s.isdecimal():
                    seconds = 3600 * int(h) + 60 * int(m) + int(s)
                case [m, s, *args] if m.isdecimal() and s.isdecimal():
                    seconds = 60 * int(m) + int(s)
                case [s, *args] if s.isdecimal():
                    seconds = int(s)
                case [*args]:
                    pass
            ctx.already_read = round(seconds / sparq(options_for_effects(effects)))
            while True:
                match args:
                    case ['tor', *args]:
                        if ctx.tor:
                            raise Explicit('duplicate tor')
                        ctx.tor = True
                    case ['ignore', *args]:
                        if ctx.ignore:
                            raise Explicit('duplicate ignore')
                        ctx.ignore = True
                    case [*args]:
                        break
            self.sources.append(ctx)
